@@ -1,13 +1,21 @@
+use crate::config::Value;
 use clap::{arg, Command};
 use evdev::{AttributeSet, Device, InputEventKind, Key};
-use nix::unistd::{Group, Uid};
+use nix::{
+    sys::stat::{umask, Mode},
+    unistd::{Group, Uid},
+};
+use signal_hook::consts::signal::*;
 use signal_hook_tokio::Signals;
 use std::{
     collections::{HashMap, HashSet},
-    env, fs,
+    env,
+    error::Error,
+    fs,
+    fs::Permissions,
     io::prelude::*,
-    os::unix::net::UnixStream,
-    path::Path,
+    os::unix::{fs::PermissionsExt, net::UnixStream},
+    path::{Path, PathBuf},
     process::{exit, id},
 };
 use sysinfo::{ProcessExt, System, SystemExt};
@@ -16,9 +24,8 @@ use tokio::time::Duration;
 use tokio::time::{sleep, Instant};
 use tokio_stream::{StreamExt, StreamMap};
 
-use signal_hook::consts::signal::*;
-
 mod config;
+mod perms;
 mod uinput;
 
 #[cfg(test)]
@@ -36,7 +43,7 @@ impl KeyboardState {
 }
 
 #[tokio::main]
-async fn main() -> Result<(), Box<dyn std::error::Error>> {
+async fn main() -> Result<(), Box<dyn Error>> {
     let args = set_command_line_args().get_matches();
     env::set_var("RUST_LOG", "swhkd=warn");
 
@@ -47,42 +54,29 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
     env_logger::init();
     log::trace!("Logger initialized.");
 
-    let pidfile: String = String::from("/tmp/swhkd.pid");
-    if Path::new(&pidfile).exists() {
-        log::trace!("Reading {} file and checking for running instances.", pidfile);
-        let swhkd_pid = match fs::read_to_string(&pidfile) {
-            Ok(swhkd_pid) => swhkd_pid,
-            Err(e) => {
-                log::error!("Unable to read {} to check all running instances", e);
-                exit(1);
-            }
-        };
-        log::debug!("Previous PID: {}", swhkd_pid);
-
-        let mut sys = System::new_all();
-        sys.refresh_all();
-        for (pid, process) in sys.processes() {
-            if pid.to_string() == swhkd_pid && process.exe() == env::current_exe().unwrap() {
-                log::error!("Swhkd is already running!");
-                exit(1);
-            }
+    let invoking_uid = match env::var("PKEXEC_UID") {
+        Ok(uid) => {
+            let uid = uid.parse::<u32>().unwrap();
+            log::trace!("Invoking UID: {}", uid);
+            uid
         }
-    }
-
-    match fs::write(&pidfile, id().to_string()) {
-        Ok(_) => {}
-        Err(e) => {
-            log::error!("Unable to write to {}: {}", pidfile, e);
+        Err(_) => {
+            log::error!("Failed to launch swhkd!!!");
+            log::error!("Make sure to launch the binary with pkexec.");
             exit(1);
         }
-    }
+    };
 
-    if let Err(_) = check_user_permissions() {
-        exit(1);
-    }
+    setup_swhkd(invoking_uid);
+
+    let root_resuid = perms::get_resuid();
+    let root_resgid = perms::get_resgid();
 
     let load_config = || {
-        let config_file_path: std::path::PathBuf = if args.is_present("config") {
+        // Drop privileges to the invoking user.
+        perms::drop_privileges(invoking_uid);
+
+        let config_file_path: PathBuf = if args.is_present("config") {
             Path::new(args.value_of("config").unwrap()).to_path_buf()
         } else {
             fetch_xdg_config_path()
@@ -90,12 +84,7 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
 
         log::debug!("Using config file path: {:#?}", config_file_path);
 
-        if !config_file_path.exists() {
-            log::error!("{:#?} doesn't exist", config_file_path);
-            exit(1);
-        }
-
-        let hotkeys = match config::load(config_file_path) {
+        let hotkeys = match config::load(&config_file_path) {
             Err(e) => {
                 log::error!("Config Error: {}", e);
                 exit(1);
@@ -112,6 +101,9 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
 
     let mut hotkeys = load_config();
 
+    // Escalate back to the root user after reading the config file.
+    perms::raise_privileges(root_resgid, root_resuid);
+
     log::trace!("Attempting to find all keyboard file descriptors.");
     let keyboard_devices: Vec<Device> =
         evdev::enumerate().filter(check_device_is_keyboard).collect();
@@ -131,8 +123,6 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
     log::debug!("{} Keyboard device(s) detected.", keyboard_devices.len());
 
     let modifiers_map: HashMap<Key, config::Modifier> = HashMap::from([
-        (Key::KEY_LEFTMETA, config::Modifier::Super),
-        (Key::KEY_RIGHTMETA, config::Modifier::Super),
         (Key::KEY_LEFTMETA, config::Modifier::Super),
         (Key::KEY_RIGHTMETA, config::Modifier::Super),
         (Key::KEY_LEFTALT, config::Modifier::Alt),
@@ -156,6 +146,7 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
 
     let mut execution_is_paused = false;
     let mut last_hotkey: Option<config::Hotkey> = None;
+    let mut pending_release: bool = false;
     let mut keyboard_states: Vec<KeyboardState> = Vec::new();
     let mut keyboard_stream_map = StreamMap::new();
 
@@ -169,11 +160,16 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
     let hotkey_repeat_timer = sleep(Duration::from_millis(0));
     tokio::pin!(hotkey_repeat_timer);
 
+    // The socket that we're sending the commands to.
+    let socket_file_path = fetch_xdg_runtime_socket_path();
     loop {
         select! {
             _ = &mut hotkey_repeat_timer, if &last_hotkey.is_some() => {
                 let hotkey = last_hotkey.clone().unwrap();
-                send_command(hotkey.clone());
+                if hotkey.keybinding.on_release {
+                    continue;
+                }
+                send_command(hotkey.clone(), &socket_file_path);
                 hotkey_repeat_timer.as_mut().reset(Instant::now() + Duration::from_millis(repeat_cooldown_duration));
             }
 
@@ -209,7 +205,9 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
                         for mut device in evdev::enumerate().filter(check_device_is_keyboard) {
                             let _ = device.ungrab();
                         }
-                        log::warn!("Got signal: {:#?}", signal);
+
+                        log::warn!("Received signal: {:#?}", signal);
+                        log::warn!("Exiting...");
                         exit(1);
                     }
                 }
@@ -235,16 +233,21 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
 
                     // Key release
                     0 => {
+                        if last_hotkey.is_some() && pending_release {
+                            pending_release = false;
+                            send_command(last_hotkey.clone().unwrap(), &socket_file_path);
+                            last_hotkey = None;
+                        }
                         if let Some(modifier) = modifiers_map.get(&key) {
                             if let Some(hotkey) = &last_hotkey {
-                                if hotkey.modifiers.contains(modifier) {
+                                if hotkey.modifiers().contains(modifier) {
                                     last_hotkey = None;
                                 }
                             }
                             keyboard_state.state_modifiers.remove(modifier);
                         } else if keyboard_state.state_keysyms.contains(key) {
                             if let Some(hotkey) = &last_hotkey {
-                                if key == hotkey.keysym {
+                                if key == hotkey.keysym() {
                                     last_hotkey = None;
                                 }
                             }
@@ -256,18 +259,18 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
                 }
 
                 let possible_hotkeys: Vec<&config::Hotkey> = hotkeys.iter()
-                    .filter(|hotkey| hotkey.modifiers.len() == keyboard_state.state_modifiers.len())
+                    .filter(|hotkey| hotkey.modifiers().len() == keyboard_state.state_modifiers.len())
                     .collect();
 
                 let event_in_hotkeys = hotkeys.iter().any(|hotkey| {
-                    hotkey.keysym.code() == event.code() &&
+                    hotkey.keysym().code() == event.code() &&
                     keyboard_state.state_modifiers
                         .iter()
-                        .all(|x| hotkey.modifiers.contains(x)) &&
-                    keyboard_state.state_modifiers.len() == hotkey.modifiers.len()
+                        .all(|x| hotkey.modifiers().contains(x)) &&
+                    keyboard_state.state_modifiers.len() == hotkey.modifiers().len()
+                    && !hotkey.is_send()
                         });
 
-
                 // Don't emit event to virtual device if it's from a valid hotkey
                 if !event_in_hotkeys {
                     uinput_device.emit(&[event]).unwrap();
@@ -282,12 +285,18 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
                 log::debug!("hotkey: {:#?}", possible_hotkeys);
 
                 for hotkey in possible_hotkeys {
-                    if keyboard_state.state_modifiers.iter().all(|x| hotkey.modifiers.contains(x))
-                        && keyboard_state.state_modifiers.len() == hotkey.modifiers.len()
-                        && keyboard_state.state_keysyms.contains(hotkey.keysym)
+                    // this should check if state_modifiers and hotkey.modifiers have the same elements
+                    if keyboard_state.state_modifiers.iter().all(|x| hotkey.modifiers().contains(x))
+                        && keyboard_state.state_modifiers.len() == hotkey.modifiers().len()
+                        && keyboard_state.state_keysyms.contains(hotkey.keysym())
                     {
                         last_hotkey = Some(hotkey.clone());
-                        send_command(hotkey.clone());
+                        if pending_release { break; }
+                        if hotkey.is_on_release() {
+                            pending_release = true;
+                            break;
+                        }
+                        send_command(hotkey.clone(), &socket_file_path);
                         hotkey_repeat_timer.as_mut().reset(Instant::now() + Duration::from_millis(repeat_cooldown_duration));
                         break;
                     }
@@ -297,22 +306,22 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
     }
 }
 
-fn sock_send(command: &str) -> std::io::Result<()> {
-    let mut stream = UnixStream::connect("/tmp/swhkd.sock")?;
+fn socket_write(command: &str, socket_path: PathBuf) -> std::io::Result<()> {
+    let mut stream = UnixStream::connect(socket_path)?;
     stream.write_all(command.as_bytes())?;
     Ok(())
 }
 
-fn send_command(hotkey: config::Hotkey) {
+fn send_command(hotkey: config::Hotkey, socket_path: &Path) {
     log::info!("Hotkey pressed: {:#?}", hotkey);
-    if let Err(e) = sock_send(&hotkey.command) {
-        log::error!("Failed to send command over IPC.");
-        log::error!("Is swhks running?");
-        log::error!("{:#?}", e)
+    if let Err(e) = socket_write(&hotkey.command, socket_path.to_path_buf()) {
+        log::error!("Failed to send command to swhks through IPC.");
+        log::error!("Please make sure that swhks is running.");
+        log::error!("Err: {:#?}", e)
     }
 }
 
-pub fn check_user_permissions() -> Result<(), ()> {
+pub fn check_input_group() -> Result<(), Box<dyn std::error::Error>> {
     if !Uid::current().is_root() {
         let groups = nix::unistd::getgroups();
         for (_, groups) in groups.iter().enumerate() {
@@ -325,7 +334,7 @@ pub fn check_user_permissions() -> Result<(), ()> {
             }
         }
         log::error!("Consider using `pkexec swhkd ...`");
-        Err(())
+        exit(1);
     } else {
         log::warn!("Running swhkd as root!");
         Ok(())
@@ -337,10 +346,10 @@ pub fn check_device_is_keyboard(device: &Device) -> bool {
         if device.name() == Some("swhkd virtual output") {
             return false;
         }
-        log::debug!("{} is a keyboard.", device.name().unwrap(),);
+        log::debug!("Keyboard: {}", device.name().unwrap(),);
         true
     } else {
-        log::trace!("{} is not a keyboard.", device.name().unwrap(),);
+        log::trace!("Other: {}", device.name().unwrap(),);
         false
     }
 }
@@ -366,8 +375,8 @@ pub fn set_command_line_args() -> Command<'static> {
     app
 }
 
-pub fn fetch_xdg_config_path() -> std::path::PathBuf {
-    let config_file_path: std::path::PathBuf = match env::var("XDG_CONFIG_HOME") {
+pub fn fetch_xdg_config_path() -> PathBuf {
+    let config_file_path: PathBuf = match env::var("XDG_CONFIG_HOME") {
         Ok(val) => {
             log::debug!("XDG_CONFIG_HOME exists: {:#?}", val);
             Path::new(&val).join("swhkd/swhkdrc")
@@ -379,3 +388,87 @@ pub fn fetch_xdg_config_path() -> std::path::PathBuf {
     };
     config_file_path
 }
+
+pub fn fetch_xdg_runtime_socket_path() -> PathBuf {
+    match env::var("XDG_RUNTIME_DIR") {
+        Ok(val) => {
+            log::debug!("XDG_RUNTIME_DIR exists: {:#?}", val);
+            Path::new(&val).join("swhkd.sock")
+        }
+        Err(_) => {
+            log::error!("XDG_RUNTIME_DIR has not been set.");
+            Path::new(&format!("/run/user/{}/swhkd.sock", env::var("PKEXEC_UID").unwrap()))
+                .to_path_buf()
+        }
+    }
+}
+
+pub fn setup_swhkd(invoking_uid: u32) {
+    // Set a sane process umask.
+    log::trace!("Setting process umask.");
+    umask(Mode::S_IWGRP | Mode::S_IWOTH);
+
+    // Get the runtime path and create it if needed.
+    let runtime_path: String = match env::var("XDG_RUNTIME_DIR") {
+        Ok(runtime_path) => {
+            log::debug!("XDG_RUNTIME_DIR exists: {:#?}", runtime_path);
+            Path::new(&runtime_path).join("swhkd").to_str().unwrap().to_owned()
+        }
+        Err(_) => {
+            log::error!("XDG_RUNTIME_DIR has not been set.");
+            String::from("/run/swhkd/")
+        }
+    };
+    if !Path::new(&runtime_path).exists() {
+        match fs::create_dir_all(Path::new(&runtime_path)) {
+            Ok(_) => {
+                log::debug!("Created runtime directory.");
+                match fs::set_permissions(Path::new(&runtime_path), Permissions::from_mode(0o600)) {
+                    Ok(_) => log::debug!("Set runtime directory to readonly."),
+                    Err(e) => log::error!("Failed to set runtime directory to readonly: {}", e),
+                }
+            }
+            Err(e) => log::error!("Failed to create runtime directory: {}", e),
+        }
+    }
+
+    // Get the PID file path for instance tracking.
+    let pidfile: String = format!("{}swhkd_{}.pid", runtime_path, invoking_uid);
+    if Path::new(&pidfile).exists() {
+        log::trace!("Reading {} file and checking for running instances.", pidfile);
+        let swhkd_pid = match fs::read_to_string(&pidfile) {
+            Ok(swhkd_pid) => swhkd_pid,
+            Err(e) => {
+                log::error!("Unable to read {} to check all running instances", e);
+                exit(1);
+            }
+        };
+        log::debug!("Previous PID: {}", swhkd_pid);
+
+        // Check if swhkd is already running!
+        let mut sys = System::new_all();
+        sys.refresh_all();
+        for (pid, process) in sys.processes() {
+            if pid.to_string() == swhkd_pid && process.exe() == env::current_exe().unwrap() {
+                log::error!("Swhkd is already running!");
+                log::error!("pid of existing swhkd process: {}", pid.to_string());
+                log::error!("To close the existing swhkd process, run `sudo killall swhkd`");
+                exit(1);
+            }
+        }
+    }
+
+    // Write to the pid file.
+    match fs::write(&pidfile, id().to_string()) {
+        Ok(_) => {}
+        Err(e) => {
+            log::error!("Unable to write to {}: {}", pidfile, e);
+            exit(1);
+        }
+    }
+
+    // Check if the user is in input group.
+    if check_input_group().is_err() {
+        exit(1);
+    }
+}
